import json
import os
from pathlib import Path


def get_json_config(config_file_name: str) -> dict:
    """
    Given a configuration file's name, read it's JSON and returns it as a dictionary.
    """
    PATH = (Path(__file__) / "../").resolve()
    file_name = os.path.join(PATH, config_file_name)
    with open(file_name) as f:
        json_config = json.loads(f.read())
    return json_config


def cs_to_list(cs_list: str = "") -> list:
    """
    converts a comma-separated string into a list of values.
    """
    if cs_list == "":
        return []
    return cs_list.replace(" ", "").split(",")


def transform_to_exclude_cs(include_cs_list: str = "", config_file_name: str = "config_hourly.json") -> list:
    """
    convert the include entities list to the exclude entities list,
    by specifying all the elements that not include in the include entities list
    """
    include_list = cs_to_list(include_cs_list)
    all_entities_list = list(get_json_config(config_file_name).keys())
    exclude_list = [entity for entity in all_entities_list if entity not in include_list]
    exclude_cs = ','.join(exclude_list)
    return exclude_cs


def get_interval_name(dag_id: str) -> str:
    """
    returns a string indicating the interval of a DAG based on its ID.
    """
    interval = 'daily' if dag_id == 'cs-cdc-to-dwh-daily' else 'hourly'
    return interval


def get_config_file_name(entity: str) -> str:
    """
    returns a string indicating the config file name that includes the given entity.
    """
    if entity in get_json_config('config_daily.json').keys():
        config_file_name = 'config_daily.json'
    elif entity in get_json_config('config_hourly.json').keys():
        config_file_name = 'config_hourly.json'
    else:
        raise ValueError(f'#### Configuration file containing the given entity ("{entity}") is not found')
    return config_file_name


def get_user_config(dag_run, ti) -> dict:
    """
    Parses a JSON string containing user input configuration, validate it, enrich it, push all the keys as Xcoms,
    and returns a dict contains the enriched user input and cluster secrets.
    """
    from airflow.models import Variable
    from airflow.exceptions import AirflowFailException

    user_config = dag_run.conf
    # validate user input
    if not user_config:
        raise AirflowFailException("#### No user input configuration although is required ####")
    for inp in ['process', 'entity', 'cluster', 'recovery_start_ts', 'recovery_end_ts']:
        try:
            user_config[inp]
        except KeyError as err:
            raise AirflowFailException(f"#### The key '{inp}' doesn't exist in the input although it's required.\n"
                                       f"error:{err} ####")
    # enrich user input
    secret = Variable.get('cs_dwh_streaming_' + user_config['cluster'] + '_secret', deserialize_json=True)
    user_config.update(secret)
    # push user input as xcom
    for key in user_config.keys():
        ti.xcom_push(key=key, value=user_config[key])
    return user_config


def get_process_task_name(dag_run) -> str:
    """
    Returns the name of the task to be executed based on the process value passed in the trigger variables of a DAG run
    """
    from airflow.exceptions import AirflowFailException

    process = dag_run.conf['process']
    if process == 'history':
        task_name = 'read_kafka_topic_history'
    elif process == 'recovery':
        task_name = 'read_kafka_topic_recovery'
    else:
        raise AirflowFailException("#### 'process' key in the trigger variables must be 'history' or 'recovery',"
                                   f"found '{process}'####")
    print(f'#### starting {process} process ####')
    return task_name


def get_member_id(user_name: str, slack_token: str) -> str:
    """
    Returns the Slack ID of a member based on their user name
    """
    from airflow.exceptions import AirflowFailException
    import requests

    try:
        user_name = user_name.strip('@wix.com')
        slack_call = f'https://slack.com/api/users.lookupByEmail?token={slack_token}&email={user_name}@wix.com'
        answer = requests.get(slack_call).json()
        member_id = answer['user']['id']
    except KeyError as e:
        raise AirflowFailException(f"#### provided user name is not in the found.\ngot '{user_name}'.\nerror :{e} ####")
    return member_id


def get_ask_approval_message(entity: str, user_name: str, recovery_table: str, slack_token: str, ti) -> None:
    """
    Sends a message to streaming channel asking for user approval to insert the recovery_table table
    into production tables. It uses the get_member_id() function to get the Slack ID and sends a message with tag.
    The message contains a unique ok_message that is pushed to XCom to validate the approval.
    """
    from uuid import uuid4

    member_id = get_member_id(user_name, slack_token)
    process_id = str(uuid4())
    ok_message = f'ok {process_id}'
    message = \
        f"<@{member_id}> Please approve the Insert of the table `sandbox.cs.{recovery_table.format(entity)}` into `{entity}` " \
        f"streaming and full table by sending the exact message in this channel -> \n \n{ok_message}"
    send_slack_message(message)
    ti.xcom_push(key='ok_message', value=ok_message)


def send_slack_message(message: str) -> None:
    """
    sends a message to a Slack channel using the SlackAPIPostOperator
    """
    from airflow.providers.slack.operators.slack import SlackAPIPostOperator

    SlackAPIPostOperator(
        task_id='ask_for_user_approval',
        slack_conn_id='slack_bot_token',
        channel='cc-data-streaming',
        text=message
    ).execute()


def get_slack_response(slack_token: str) -> str:
    """
    returns the most recent message in the "cc-data-streaming" Slack channel, converted to lowercase
    """
    import requests

    slack_call = 'https://slack.com/api/conversations.history?channel=C03N27UE945&limit=1' \
                 f'&token={slack_token}'
    res = requests.get(slack_call).json()
    response = res['messages'][0]['text']
    return response.lower()


def get_user_approval(user_name: str, ok_message: str, slack_token: str) -> bool:
    """
    Waits for a user response in the "cc-data-streaming" Slack channel for a maximum of 10 minutes.
    If the user responds with the ok_message that was sent in get_ask_approval_message(), this function returns True.
    If the user does not respond (or respond with wrong input) within the time limit, an exception is raised.
    """
    from airflow.exceptions import AirflowFailException
    from time import sleep

    min_to_wait = 10
    while min_to_wait > 0:
        answer = get_slack_response(slack_token)
        if answer == ok_message:
            print(f'#### user {user_name} has approved the insert. got the answer: {answer} ####')
            return True
        else:
            print(f'#### waiting for user response ####')
            sleep(60)
            min_to_wait -= 1
    raise AirflowFailException(f'#### user {user_name} has NOT approved the insert. got the answer: {answer} '
                               f'expect {ok_message} ####')


def create_kafka_table_if_not_exists(catalog: str, schema: str, history_table: str) -> None:
    """
    Creates a table in Trino with the specified name and schema if it doesn't already exist.
    Raise Error if the table already exists.
    """
    from wix_trino_client.trino_connection import WixTrinoConnection, DatabaseError

    pc = WixTrinoConnection()
    try:
        table_creation_statement = \
            f""" CREATE TABLE {catalog}.{schema}.{history_table}(
                    before varchar,
                    after varchar,
                    batch_id varchar,
                    process_timestamp timestamp(6),
                    kafka_key varchar,
                    kafka_topic varchar,
                    kafka_partition integer,
                    kafka_offset bigint,
                    kafka_timestamp timestamp(6),
                    kafka_timestamptype integer
                    )"""
        pc.execute_sql(table_creation_statement)
    except DatabaseError as e:
        raise Exception(f'#### failed to create the table {catalog}.{schema}.{history_table}, '
                        f'overwrite existing table is not allowed.\n'
                        f'query:\nby executing:\n{table_creation_statement} \ngot the error:{e}####')
    print(f'#### created the table {catalog}.{schema}.{history_table}\nby executing:\n{table_creation_statement} ####')


def create_s3_folder_if_not_exists(s3_full_path: str) -> None:
    """
    Creates a folder in S3 bucket in the entity name, if it doesn't already exist.
    Raise Error if the folder already exists.
    """
    import boto3

    s3_resources = boto3.resource('s3')
    s3_client = boto3.client('s3')
    bucket_name, cluster_folder_name, entity_folder_name = s3_full_path.strip('s3://').split('/')
    bucket = s3_resources.Bucket(bucket_name)
    full_folder_name = cluster_folder_name + '/' + entity_folder_name + '/'

    # check if folder already exists
    for obj in bucket.objects.all():
        if full_folder_name in obj.key:
            raise Exception(f'#### {full_folder_name} already exists in {bucket_name} bucket. ####')

    s3_client.put_object(Bucket=bucket_name, Key=full_folder_name)
    print(f'#### {full_folder_name} created successfully in {bucket_name} bucket. ####')


def get_table_batch_records(catalog: str, schema: str, table: str) -> list:
    """
    Querying the num of records fetched from given table.
    """
    from wix_trino_client.trino_connection import WixTrinoConnection

    pc = WixTrinoConnection()
    res = pc.execute_sql(f'select batch_id, count(*) as num_of_records '
                         f'from {catalog}.{schema}.{table} '
                         f'group by batch_id '
                         f'order by CAST(batch_id AS INT) desc '
                         f'limit 1')
    return res


def check_if_process_succeeded(catalog: str, schema: str, table: str) -> None:
    """
    Checking if the history process was successful by querying the num of records fetched
    in the last streaming batch. fails the task if the table empty or doesn't exists.
    """
    from wix_trino_client.trino_connection import DatabaseError
    from airflow.exceptions import AirflowFailException

    try:
        res = get_table_batch_records(catalog, schema, table)
        if res[0][1] == 0:
            raise AirflowFailException(f'#### process fail, try again ####')
    except DatabaseError as e:
        raise AirflowFailException(f'#### process fail, try again ####')
# TODO: move here the non-spark func from



from setuptools import setup

setup(
    name='cc_dwh_streaming_pkg',
    version='1.0.1',
    description='serving CC dwh streaming apps',
    author='Asaf Shelach',
    packages=['cc_dwh_streaming_pkg'],
    install_requires=[
        'pyspark.sql',
        'argparse'
    ],
)



from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import StructField, StructType, StringType
import pyspark.sql.functions as F
from pyspark.sql.window import Window
import argparse
from cs.dwh_streaming.share.cc_dwh_streaming_pkg.py_utils import cs_to_list, get_table_batch_records


class KeyValue(argparse.Action):
    """
    Custom action class for argument parsing that takes a list of
    key-value pairs and stores them as a dictionary.
    """

    def __call__(self, parser, namespace,
                 values, option_string=None):
        setattr(namespace, self.dest, dict())
        for value in values:
            # split it into key and value
            key, value = value.split('=')
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


def get_cmd_vars(args: dict) -> dict:
    """
    Parses command line arguments into a dictionary.
    """
    parser = argparse.ArgumentParser()
    for arg in args['required']:
        parser.add_argument(f"--{arg}", required=True)
    for arg in args['optional']:
        parser.add_argument(f"--{arg}")
    parser.add_argument("--kwargs", nargs='*', action=KeyValue)
    args = parser.parse_args()
    args_dict = vars(args)
    if args.kwargs:
        for key in args.kwargs:
            args_dict.setdefault(key, args.kwargs[key])
    args_dict.pop('kwargs', None)
    return args_dict


def get_message_schema(cluster: str = 'answers') -> StructType:
    """
    Returns the schema for the CDC message as StructType object.
    """
    if cluster == 'answers' or cluster == 'answers_dev':
        payload_schema = [
            StructField("before", StringType()),
            StructField("after", StringType()),
            StructField("source", StringType()),
            StructField("op", StringType()),
            StructField("ts_ms", StringType()),
            StructField("transaction", StringType())
        ]

    else:
        payload_schema = [
            StructField("data", StringType()),
            StructField("source", StringType())
        ]
    message_schema = (
        StructType([
            StructField("schema", StringType()),
            StructField("payload", StructType(payload_schema))
        ]))
    return message_schema


def get_kafka_config(cluster_api_key_var: str, cluster_api_secret_var: str, kafka_host_var: str,
                     kafka_server_var: str) -> dict:
    """
    Returns a dictionary containing Kafka configuration settings.
    """
    cluster_api_key = cluster_api_key_var
    cluster_api_secret = cluster_api_secret_var
    kafka_host = kafka_host_var
    kafka_server = kafka_server_var
    kafka_config = {
        "kafka_sasl_jaas_config": f'org.apache.kafka.common.security.plain.PlainLoginModule required username="{cluster_api_key}" password="{cluster_api_secret}";',
        "kafka.sasl.mechanism": "PLAIN",
        "kafka.security.protocol": "SASL_SSL",
        "kafka_bootstrap_servers": f"{kafka_host}:{kafka_server}",
        "client.dns.lookup": "use_all_dns_ips",
        "minPartitions": 300
    }
    return kafka_config


def generate_topics_dict(topic_entities: dict) -> dict:
    """
    Given a dictionary of entities' topics, generates a new dictionary mapping each entity to a corresponding topic name.
    """
    topics_dict = {
        entity: f"cdc.cc-dwh-{topic_entities[entity]['entity_name']}.answers.wix_connect.{topic_entities[entity]['entity_table']}.binlog"
        for entity in topic_entities}
    return topics_dict


def parsed_raw_messages(df: DataFrame, cluster: str = 'answers') -> DataFrame:
    """
    Parses DataFrame of raw messages from a Kafka topic using a predefined schema and returns a DataFrame.
    """
    schema = get_message_schema(cluster)
    parsed_df = df \
        .selectExpr("cast (value as STRING) as jsonData", "cast(key as STRING) as kafka_key", "topic as kafka_topic",
                    "partition as kafka_partition", "offset as kafka_offset", "timestamp as kafka_timestamp",
                    "timestampType as kafka_timestampType") \
        .select(F.from_json("jsonData", schema).alias("message"),
                F.current_timestamp().alias("process_timestamp"), "kafka_key", "kafka_topic",
                "kafka_partition", "kafka_offset", "kafka_timestamp", "kafka_timestampType")
    if cluster == 'answers' or cluster == 'answers_dev':
        parsed_df = parsed_df.select("message.payload.before", "message.payload.after", "process_timestamp",
                                     "kafka_key", "kafka_topic", "kafka_partition", "kafka_offset", "kafka_timestamp",
                                     "kafka_timestampType")
    else:
        parsed_df = parsed_df.select("message.payload.data", "process_timestamp", "kafka_key", "kafka_topic",
                                     "kafka_partition", "kafka_offset", "kafka_timestamp", "kafka_timestampType")
    return parsed_df


def writeStreamforeachBatch(batch_df, batch_id, spark, query_runner, catalog, schema, stream_table, full_table):
    """
    Writes a batch of data from a streaming DataFrame to two presto tables.
    """
    batch_df = batch_df.withColumn('batch_id', F.lit(batch_id)) \
        .select("before", "after", "batch_id", "process_timestamp",
                "kafka_key", "kafka_topic", "kafka_partition",
                "kafka_offset", "kafka_timestamp", "kafka_timestampType")
    batch_df.persist()
    spark.catalog.dropGlobalTempView("stream_temp")
    batch_df.createGlobalTempView("stream_temp")
    storing_tables = [f'{full_table}', f'{stream_table}']
    for table in storing_tables:
        if table == '':
            continue
        print(f"#### storing data in {catalog}.{schema}.{table} #####")  # TODO: logging
        query_runner.sql(
            f"""INSERT INTO {catalog}.{schema}.{table}
                    SELECT *
                    FROM global_temp.stream_temp""")  # TODO: consider ParquetTableStorer/df.write instead
    batch_df.unpersist()


def process_batch_wrapper(offsets_per_trigger: int, **kwargs) -> None:
    """
    Wrapper for writeStreamforeachBatch func
    that also check if we fetched less than the max offset and stopping the stream
    """
    writeStreamforeachBatch(**kwargs)
    res = get_table_batch_records(kwargs["catalog"], kwargs["schema"], kwargs["full_table"])
    try:
        num_of_records = res[0][1]
    except IndexError:
        num_of_records = 0
    if num_of_records > (offsets_per_trigger / 2):
        print(f'#### found {num_of_records} records, continuing ####')
    else:
        raise Exception(f'#### found {num_of_records} records, ending the stream ####')


def prepared_stream_df(stream_df: DataFrame) -> DataFrame:
    """
    prepares a given streaming DataFrame of CDC records,by selecting specific columns
    and generating a "record_type" column based on the before and after values.
    returns the prepared DataFrame.
    """
    stream_df_prep = stream_df \
        .select('before', 'after', 'kafka_timestamp', 'kafka_offset',
                F.lit(F.when((F.col('before').isNotNull() & F.col('after').isNotNull()), 'updated')
                      .when((F.col('before').isNotNull() & F.col('after').isNull()), 'deleted')
                      .when((F.col('after').isNotNull() & F.col('before').isNull()), 'created')
                      .otherwise('non-cdc')).alias('record_type')) \
        .filter(F.col('record_type') != 'non-cdc') \
        .select('record_type',
                F.lit(
                    F.when(((F.col('record_type') == 'updated') | (F.col('record_type') == 'created')), F.col('after'))
                    .otherwise(F.col('before'))).alias('payload')
                , F.lit(
            F.when(((F.col('record_type') == 'updated') | (F.col('record_type') == 'created')), F.col('before'))
            .otherwise(F.col('after'))).alias('prev_payload')
                , F.lit(F.when((F.col('record_type') == 'deleted'), 1)
                        .otherwise(0)).alias('is_hard_deleted')
                , 'kafka_timestamp', 'kafka_offset'
                )
    return stream_df_prep


def generate_payload_expr(payload_columns: list) -> list:
    """
    Given a list of column names, generates a list of JSON expressions used to extract values from a payload column.
    """
    payload_expr = [F.get_json_object('payload', f'$.{column}').alias(f'{column}') for column in payload_columns]
    return payload_expr


def generate_coalesce_expr(coalesce_columns: dict) -> list:
    """
    Given a dictionary mapping coalesce column names to a new columns, generates a list of
    expressions that coalesce each column and its alternative(s) into a new column with the given column name.
    """
    coalesce_expr = [F.coalesce(*[F.col(column) for column in cs_to_list(columns)])
                         .alias(coalesce_columns[columns]) for columns in coalesce_columns]
    return coalesce_expr


def generate_nested_expr(nested_columns: dict) -> list:
    """
    Given a dictionary mapping column names to a list of sub-column names, generates a nested list of expressions used
    to extract values from a JSON column.
    """
    nested_expr = [[F.get_json_object(column, f'$.{sub_column}').alias(f'{nested_columns[column][sub_column]}')
                    for sub_column in nested_columns[column]] for column in nested_columns]
    return nested_expr


def generate_casting_item(column: str, cast_to: str, cast_from: str = 'string'):
    """
    Given a column name, target type, and (optionally) source type, generates an expression used to cast the column
    to the target type.
    """
    if cast_to == 'timestamp':
        return F.to_timestamp(F.col(column) / 1000).alias(column)  # TODO: other types
    if cast_to == 'array<string>':
        return F.from_json(F.col(column), cast_to).alias(column)
    if cast_to == 'boolean':
        return F.col(column).cast("boolean").alias(column)


def generate_casting_list(df_columns: list, casting_column: dict) -> list:
    """
    Generates a list of PySpark expressions for casting the columns of a DataFrame based on a dictionary mapping column
    names to target data types.
    """
    casting_expr = [(generate_casting_item(df_col, casting_column[df_col])
                     if df_col in casting_column.keys() else F.col(df_col)) for df_col in df_columns]
    return casting_expr


def df_raw_to_slowly(raw_df: DataFrame, key_col: str = 'id', order_by_col: str = 'last_update_date') -> DataFrame:
    """
    Given a raw DataFrame, key columns and the order-by columns, transforms a raw dataframe into a slowly changing
    dimension (SCD) Type 2 dataframe. It adds a "rev_in_batch" column that contains the revision number for each record,
    based on the order of the specified key columns and the specified order-by columns (descending),
    as well as the Kafka timestamp and offset. The function returns the transformed dataframe.
    """
    key_col_list, order_by_col_list = cs_to_list(key_col), cs_to_list(order_by_col)
    key_expr = [F.col(key) for key in key_col_list]
    order_by_expr = [F.col(order_key).desc() for order_key in order_by_col_list + ['kafka_timestamp', 'kafka_offset']]
    partition = Window.partitionBy(key_expr).orderBy(order_by_expr)
    slowly_df = raw_df \
        .withColumn('rev_in_batch', F.row_number().over(partition))
    return slowly_df


def df_slowly_to_latest(slowly_df: DataFrame) -> DataFrame:
    """
    Filters out all but the latest revision of each record in a slowly changing dimension (SCD) dataframe
    and returns the filtered dataframe.
    """
    latest_df = slowly_df.filter(F.col('rev_in_batch') == 1) \
        .select([df_col for df_col in slowly_df.columns if df_col != 'rev_in_batch'])
    return latest_df

# common TODO: 1.consider ParquetTableStorer vs. spark.sql commands
#              2. df.repartition('device_type').sortWithinPartitions('device_type') when writing the tables
